#include <process.hpp>
#include <iostream>
#include <sstream>
#include "ProcessEngineCommand.h"
#include "config.h"

ProcessEngineCommand::ProcessEngineCommand(const ProcessManager* manager, std::string command)
  : manager(manager), command(std::move(command)), executed(false), jsonParsed(false) {

}

void ProcessEngineCommand::fillInParameter(std::string value) {
  parameterValues.push_back(std::move(value));
}

std::string ProcessEngineCommand::getTextResult() {
  return stdoutResult;
}

json ProcessEngineCommand::getJsonResult() {
  if (!executed) {
    execute();
  }
  if (!jsonParsed) {
    parseJson();
  }

  return jsonResult;
}

int ProcessEngineCommand::getExitCode() {
  if (!executed) {
    execute();
  }
  return exitCode;
}

void ProcessEngineCommand::execute() {
  executed = true;

  std::string parameterizedCommand = command;
  for (const auto &value : parameterValues) {
    auto tokenPosition = parameterizedCommand.find(config::PAR_REPLACE_TOKEN);
    if (tokenPosition != std::string::npos) {
      parameterizedCommand.replace(tokenPosition, config::PAR_REPLACE_TOKEN.size(), value);
    } else {
      manager->log("Unable to parameterize command " + command + " with value " + value);
    }
  }

  std::string processEngineStdout, processEngineStderr;
  TinyProcessLib::Process runProcess(parameterizedCommand, "",
     [&processEngineStdout](const char *bytes, size_t n) mutable {
       processEngineStdout.append(std::string(bytes, n));
     }, [&processEngineStderr](const char *bytes, size_t n) mutable {
      processEngineStderr.append(std::string(bytes, n));
    });
  int exit_status;
  while (!runProcess.try_get_exit_status(exit_status)) {
    std::this_thread::sleep_for(std::chrono::microseconds(100));
  }
  // note: the following lines only work _after_ the while loop finished (otherwise the process
  //  might still be running)
  stdoutResult = processEngineStdout;
  stderrResult = processEngineStderr;
  exitCode = exit_status;
}

std::string ProcessEngineCommand::getStdErr() {
  if (!executed) {
    execute();
  }
  return stderrResult;
}

void ProcessEngineCommand::parseJson() {
  std::string processEngineStdout = stdoutResult;
  // replace all new lines to avoid parsing errors:
  std::replace(processEngineStdout.begin(), processEngineStdout.end(), '\n', ' ');

  jsonResult = json();
  if (!processEngineStdout.empty()) {
    try {
      jsonResult = json::parse(processEngineStdout);
      if (jsonResult.empty()) {
        manager->log("Process engine returned empty json");
      }
    } catch(json::parse_error& error) {
      std::ostringstream logStream;
      logStream << "error parsing process engine response: " << error.what() << std::endl;
      manager->log(logStream.str());
      std::cerr << logStream.str();
    }
  } else {
    manager->log("Workflow Engine returned empty string. Json expected.");
  }
  jsonParsed = true;
}
