#include <fstream>
#include "ProcessManager.h"
#include <nlohmann/json.hpp>
#include <iomanip>
#include <iostream>
#include <utility>
#include <algorithm>
#include <string>
#include <process.hpp>
#include "util.h"
#include "config.h"
#include "ProcessEngineCommand.h"

using json = nlohmann::json;


ProcessManager::ProcessManager(std::string basePath, std::string workflowDirectory, std::string engineFile, std::string statusFile)
  : basePath(std::move(basePath)), workflowsDirectory(std::move(workflowDirectory)), engineFile(std::move(engineFile)),
    workflowsFile(std::move(statusFile)) {

  if (this->engineFile.empty()) {
    this->engineFile = config::DEFAULT_ENGINE_FILENAME;
  }
  if (this->workflowsFile.empty()) {
    this->workflowsFile = config::DEFAULT_WORKFLOW_FILENAME;
  }

  loadEngines();
  loadWorkflows();
}

bool ProcessManager::startWorkflow(const std::string &workflowFile, const std::string& workflowEngine) {
  std::string workflowFileAbsolute = workflowFile;
  if (!util::isAbsolutePath(workflowFile)) {
    workflowFileAbsolute = util::getAbsolutePath(workflowFile);
  }
  if (!util::fileExists(workflowFileAbsolute)) {
    error("Can not open workflow file: " + workflowFileAbsolute);
    return false;
  }
  std::string engineName = workflowEngine.empty() ? defaultProcessEngine : workflowEngine;
  const auto engine = getProcessEngine(engineName, true);

  std::string workflowFolder = getWorkflowExecutionFolder(nextId);

  if (!util::createDirectory(workflowFolder)) {
    error("Can not create directory for workflow execution: " + workflowFolder);
    return false;
  }

  if (engine->getStartCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    nextId++; // the folder was already created
    saveWorkflows();
    return false;
  }
  ProcessEngineCommand engineCommand(this, engine->getStartCommand());
  engineCommand.fillInParameter(workflowFolder);
  engineCommand.fillInParameter(workflowFileAbsolute);

  if (engineCommand.getExitCode() == 0) {
    try {
      json returnedJson = engineCommand.getJsonResult();
      auto createdWorkflow = std::make_unique<Workflow>();
      createdWorkflow->fromJson(returnedJson);
      createdWorkflow->setId(nextId);
      std::cout << std::setw(4) << createdWorkflow->getJson() << std::endl;
      workflows[createdWorkflow->getId()] = std::move(createdWorkflow);

      nextId++;
      saveWorkflows();
    } catch (json::exception &exception) {
      error("Parsing error: " + std::string(exception.what()));
      nextId++; // the folder was already created
      saveWorkflows();
      return false;
    };
  } else {
    error("Workflow was not accepted by the process engine. Error from Process Engine: "
      + engineCommand.getStdErr());
    nextId++; // the folder was already created
    saveWorkflows();
    return false;
  }
  return true;
}

bool ProcessManager::listWorkflows() const {

  json workflowsArray = json::array();
  bool workflowUpdated = false;
  for (const auto &workflow : workflows) {
    // refresh workflow info if workflow is not finished
    workflowUpdated |= updateWorkflow(workflow.second.get());

    workflowsArray.push_back(workflow.second->getJson());
  }

  if (workflowUpdated) {
    saveWorkflows();
  }

  std::cout << std::setw(4) << workflowsArray << std::endl;
  return true;
}

bool ProcessManager::updateWorkflow(Workflow* workflow) const {
  bool workflowUpdated = false;
  WorkflowState lastKnownState = workflow->getState();
  if (lastKnownState == RUNNING || lastKnownState == READY || lastKnownState == NEEDS_INTERACTION) {
    try {
      json jsonStatus = retrieveWorkflowStatus(workflow->getId());
      if (!jsonStatus.empty()) {
        workflowUpdated = true;
        if (jsonStatus.contains("id")) {
          // do not let process engines overwrite the id
          jsonStatus.erase("id");
        }
        workflow->fromJson(jsonStatus);
      }
    } catch(const std::exception& exception) {
      error(exception.what());
      return false;
    }
  }
  return workflowUpdated;
}

bool ProcessManager::listEngines() const {
  json enginesArray = json::array();
  for (const auto &engine : processEngines) {
    enginesArray.push_back(engine.second->getJson());
  }
  std::cout << std::setw(4) << enginesArray << std::endl;
  return true;
}

void ProcessManager::loadEngines() {
  std::string fileName = basePath + engineFile;

  if (!util::fileExists(fileName) && util::fileExists(config::DEFAULT_ENGINE_CONFIGURATION_PATH)) {
    // copy the default engine configuration
    if (util::copyFile(config::DEFAULT_ENGINE_CONFIGURATION_PATH, fileName)) {
      std::cerr << "Info: Using default process engine configuration" << std::endl;
    } else {
      std::cerr << "Warning: Could not fall back to default process engine configuration" << std::endl;
    }
  }

  json jsonObj;
  std::ifstream in(fileName);
  if (util::fileExists(fileName)) {
    try {
      in >> jsonObj;
    } catch(nlohmann::detail::parse_error& error) {
      std::cerr << "Warning: Unable to read process engine descriptions" << std::endl;
      std::cerr << error.what() << std::endl; // debug
      return;
    }
  }

  json engineArray = jsonObj[config::JSON_ENGINES_KEY];
  if (!engineArray.empty() && engineArray.type() == json::value_t::array) {
    for (json::iterator it = engineArray.begin(); it != engineArray.end(); ++it) {
      auto engine = std::make_unique<ProcessEngine>();
      engine->fromJson(*it);
      processEngines[engine->getName()] = std::move(engine);
    }
  } else {
    std::cerr << "Warning: No process engine descriptions found in " << fileName << std::endl;
  }
  if (jsonObj.contains(config::JSON_DEFAULT_ENGINE_KEY) && !jsonObj[config::JSON_DEFAULT_ENGINE_KEY].empty()) {
    defaultProcessEngine = jsonObj[config::JSON_DEFAULT_ENGINE_KEY];
  } else if(!processEngines.empty()) {
    std::cerr << "Warning: No default process engine specified in " << fileName << std::endl;
  }
}

void ProcessManager::loadWorkflows() {
  std::string fileName = basePath + workflowsFile;

  std::ifstream in(fileName);
  json jsonObj;
  if (util::fileExists(fileName)) {
    try {
      in >> jsonObj;
    } catch(nlohmann::detail::parse_error& error) {
      std::cout << "No workflow descriptions to restore." << std::endl;
    }
  }

  if (jsonObj.contains(config::JSON_NEXTID_KEY) && jsonObj[config::JSON_NEXTID_KEY].is_number()) {
    nextId = jsonObj[config::JSON_NEXTID_KEY].get<int>();
  } else {
    nextId = 0;
    if (!workflows.empty()) {
      throw std::logic_error("Next workflow id could not be found in non-empty persistence file");
    }
  }
  json workflowArray = jsonObj["workflows"];
  if (!workflowArray.empty() && workflowArray.type() == json::value_t::array) {
    for (json::iterator it = workflowArray.begin(); it != workflowArray.end(); ++it) {
      if (*it != nullptr) {
        auto workflow = std::make_unique<Workflow>();
        workflow->fromJson(*it);
        workflows[workflow->getId()] = std::move(workflow);
      }
    }
  }
}

void ProcessManager::saveWorkflows() const {
  std::string fileName = basePath + workflowsFile;
  std::ofstream out(fileName);
  json newObj;
  for (const auto &workflow : workflows) {
    newObj[config::JSON_WORKFLOWS_KEY].push_back(workflow.second->getJson());
  }
  newObj[config::JSON_NEXTID_KEY] = nextId;
  out << std::setw(4) << newObj << std::endl;
}

json ProcessManager::retrieveWorkflowStatus(unsigned int id) const {
  // note: can not use workflow[...] since this method is const (and unordered_map::operator[] is not)
  // -> use the iterator returned by find instead to access elements here

  // note: do not abort on errors here because it most likely will only affect some workflows
  // Do not write every error to log because this would spam the log drastically
  auto workflow = getExistingWorkflow(id, false, false);
  const ProcessEngine* engine = nullptr;
  if (workflow != nullptr) {
    engine = getProcessEngine(workflow->getProcessEngine(), false, false);
  } else {
    throw std::logic_error("Unknown workflow");
  }

  std::string workflowFolder = getWorkflowExecutionFolder(id);
  if (engine != nullptr) {
    if (engine->getStatusCommand().empty()) {
      throw std::logic_error("The responsible process engine does not support this operation, or the command was not configured.");
    }
    ProcessEngineCommand engineCommand(this, engine->getStatusCommand());
    engineCommand.fillInParameter(workflowFolder);
    if (engineCommand.getExitCode() != 0) {
      std::ostringstream logStream;
      logStream << "Process Engine returned non zero exit code when retrieving status: " << engineCommand.getStdErr();
      log(logStream.str());
    }
    return engineCommand.getJsonResult();
  } else {
    throw std::logic_error("Process Engine not found");
  }
}

std::string ProcessManager::getWorkflowExecutionFolder(unsigned int id) const {
  return basePath + workflowsDirectory + std::to_string(id) + "/";
}

bool ProcessManager::printWorkflowLog(unsigned int workflowId) {

  auto workflow = getExistingWorkflow(workflowId, true);
  auto engine = getProcessEngine(workflow->getProcessEngine(), true);


  std::string workflowFolder = getWorkflowExecutionFolder(workflowId);
  if (engine->getLogCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    return false;
  }
  ProcessEngineCommand engineCommand(this, engine->getLogCommand());
  engineCommand.fillInParameter(workflowFolder);
  if (engineCommand.getExitCode() != 0) {
    std::ostringstream logStream;
    logStream << "Process Engine returned non zero exit code when retrieving the workflow log: " << engineCommand.getStdErr();
    error(logStream.str());
    return false;
  }
  std::cout << engineCommand.getTextResult() << std::endl;
  return true;
}

void ProcessManager::log(const std::string& message) const {
  if (config::LOGGING_ENABLED) {
    std::string logFileName = basePath + config::LOGFILE;
    std::ofstream logStream;

    logStream.open(logFileName, std::ios_base::app); // append mode

    auto t = std::time(nullptr);
    auto tm = *std::localtime(&t);
    logStream << "[" << std::put_time(&tm, config::DATETIME_FORMAT.c_str()) << "] "
              << message << std::endl;
  }
}

const ProcessEngine * ProcessManager::getProcessEngine(const std::string &engineName, bool exitOnError,
                                                       bool logErrors) const {
    auto engineItem = processEngines.find(engineName);
    if (engineItem != processEngines.end()) {
      return engineItem->second.get();
    } else if (exitOnError || logErrors) {
      std::ostringstream logStream;
      logStream << "Process engine with the name " << engineName << " not found.";
      if (logErrors) {
        log(logStream.str());
      }
      if (exitOnError) {
        error(logStream.str());
        exit(1);
      }
    }
    return nullptr;
}

Workflow * ProcessManager::getExistingWorkflow(unsigned int workflowId, bool exitOnError, bool logErrors) const {
  auto workflowItem = workflows.find(workflowId);
  if (workflowItem != workflows.end()) {
    return workflowItem->second.get();
  } else if (exitOnError || logErrors) {
    std::ostringstream logStream;
    logStream << "Workflow with id " << std::to_string(workflowId) << " not found.";
    if (logErrors) {
      log(logStream.str());
    }
    if (exitOnError) {
      error(logStream.str());
      exit(1);
    }
  }
  return nullptr;
}

bool ProcessManager::printWorkflowShortcuts(unsigned int workflowId) {

  auto workflow = getExistingWorkflow(workflowId, true);
  auto engine = getProcessEngine(workflow->getProcessEngine(), true);

  if (engine->getShortcutsCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    return false;
  }
  std::string workflowFolder = getWorkflowExecutionFolder(workflowId);
  ProcessEngineCommand engineCommand(this, engine->getShortcutsCommand());
  engineCommand.fillInParameter(workflowFolder);
  if (engineCommand.getExitCode() != 0) {
    std::ostringstream logStream;
    logStream << "Process Engine returned non zero exit code when retrieving the workflow shortcuts: " << engineCommand.getStdErr();
    error(logStream.str());
    return false;
  }
  std::cout << engineCommand.getTextResult() << std::endl;
  return true;
}

bool ProcessManager::printWorkflowInteractions(unsigned int workflowId) {

  auto workflow = getExistingWorkflow(workflowId, true);
  auto engine = getProcessEngine(workflow->getProcessEngine(), true);

  std::string workflowFolder = getWorkflowExecutionFolder(workflowId);
  if (engine->getInteractionsCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    return false;
  }
  ProcessEngineCommand engineCommand(this, engine->getInteractionsCommand());
  engineCommand.fillInParameter(workflowFolder);
  if (engineCommand.getExitCode() != 0) {
    std::ostringstream logStream;
    logStream << "Process Engine returned non zero exit code when retrieving the workflow interactions: " << engineCommand.getStdErr();
    error(logStream.str());
    return false;
  }
  // TODO if the result is empty, at least print minimal json with an empty list
  std::cout << engineCommand.getTextResult() << std::endl;
  return true;
}

bool ProcessManager::continueWorkflow(unsigned int workflowId) {
  auto workflow = getExistingWorkflow(workflowId, true);
  auto engine = getProcessEngine(workflow->getProcessEngine(), true);

  std::string workflowFolder = getWorkflowExecutionFolder(workflowId);
  if (engine->getContinueCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    return false;
  }
  ProcessEngineCommand engineCommand(this, engine->getContinueCommand());
  engineCommand.fillInParameter(workflowFolder);
  engineCommand.fillInParameter(workflow->getFileName());
  if (engineCommand.getExitCode() != 0) {
    std::ostringstream logStream;
    logStream << "Process Engine returned non zero exit code when attempting to continue workflow execution: " << engineCommand.getStdErr();
    error(logStream.str());
    return false;
  }
  std::cout << engineCommand.getTextResult() << std::endl;
  return true;
}

void ProcessManager::error(const std::string& message) const {
  std::cerr << message << std::endl;
  log("Error: " + message);
}

bool ProcessManager::inputInteractionValue(unsigned int workflowId, std::string interactionId,
                                           const std::string &inputValue) {
  auto workflow = getExistingWorkflow(workflowId, true);
  auto engine = getProcessEngine(workflow->getProcessEngine(), true);

  std::string workflowFolder = getWorkflowExecutionFolder(workflowId);
  if (engine->getInputInteractionValueCommand().empty()) {
    error("The responsible process engine does not support this operation, or the command was not configured.");
    return false;
  }
  ProcessEngineCommand engineCommand(this, engine->getInputInteractionValueCommand());
  engineCommand.fillInParameter(workflowFolder);
  engineCommand.fillInParameter(std::move(interactionId));
  engineCommand.fillInParameter("\""+inputValue+"\"");
  if (engineCommand.getExitCode() != 0) {
    std::ostringstream logStream;
    logStream << "Process Engine returned non zero exit code when attempting to input interaction value: " << engineCommand.getStdErr();
    error(logStream.str());
    return false;
  }
  return true;
}

bool ProcessManager::printWorkflowStatus(unsigned int workflowId) {
  try {
    auto workflow = getExistingWorkflow(workflowId, true, true);
    if (workflow) {
      updateWorkflow(workflow);
      std::cout << std::setw(4) << workflow->getJson() << std::endl;
      return true;
    }
  } catch(const std::exception& exception) {
    error(exception.what());
    return false;
  }
  return false;
}
