#include <CLI/CLI.hpp>
#include "ProcessManager.h"
#include "util.h"


int main(int argc, char** argv) {
  CLI::App app{"Process manager: Commandline tool to manage workflow execution."};

  std::string path = getenv("HOME");
  path += "/";
  path += ".process_manager/";
  if (!util::fileExists(path) && !util::createDirectory(path)) {
    throw std::logic_error("Unable to create directory " + path);
  }

  std::string workflowFile, workflowEngine, inputValue;
  int workflowId = 0;
  std::string interactionId;

  app.add_option("-p,--path", path,
                 "Base directory to execute in (application needs write permissions here). Defaults to $HOME/.process_manager");
  //app.add_flag("-m,--machine-readable", machineReadableOutput, "Give any output in a machine readable format (json)");
  app.require_subcommand(1);

  CLI::App *listCommand = app.add_subcommand("list", "List workflows or available workflow engines");
  CLI::App *listWorkflowsCommand = listCommand->add_subcommand("workflows", "List workflows and execution history");
  CLI::App *listEnginesCommand = listCommand->add_subcommand("engines", "List available workflow engines");
  listCommand->require_subcommand(1); // one subcommand must be given

  CLI::App *statusCommand = app.add_subcommand("status", "Print status of the workflow with the given id");
  statusCommand->add_option("workflowId", workflowId, "Id of the workflow")->required();

  CLI::App *startWorkflowCommand = app.add_subcommand("start", "Start the workflow from a .flow file");
  startWorkflowCommand->add_option("file", workflowFile, "Workflow description file (.flow)")->required();
  startWorkflowCommand->add_option("-e,--engine", workflowEngine, "Workflow engine for processing. Use --list-engines for a list of available engines.");

  CLI::App *continueWorkflowCommand = app.add_subcommand("continue", "Continue execution of a workflow with given id.");

  CLI::App *getWorkflowLogCommand = app.add_subcommand("log", "Get the execution log of a workflow");
  CLI::App *getWorkflowShortcutsCommand = app.add_subcommand("shortcuts", "Get the shortcuts of a workflow");
  CLI::App *getWorkflowInteractionsCommand = app.add_subcommand("interactions", "Get interaction descriptions of a workflow");

  CLI::App *inputInteractionValueCommand = app.add_subcommand("input", "Provide input value for requested user interaction to the workflow engine");
  inputInteractionValueCommand->add_option("workflowId", workflowId, "Id of the workflow")->required();
  inputInteractionValueCommand->add_option("interactionId", interactionId, "Id of the workflow")->required();
  inputInteractionValueCommand->add_option("value", inputValue, "Value")->required();

  // add id argument to all commands which need the workflow id
  for (auto command : {continueWorkflowCommand, getWorkflowLogCommand, getWorkflowShortcutsCommand, getWorkflowInteractionsCommand}) {
    command->add_option("id", workflowId, "Id of the workflow")->required();
  }

  CLI11_PARSE(app, argc, argv);

  if (path.back() != '/') {
    path += '/';
  }

  ProcessManager manager(path, "workflows/", "engines.json", "status.json");
  bool success = true;
  if (*listWorkflowsCommand) {
    success = manager.listWorkflows();
  } else if(*listEnginesCommand) {
    success = manager.listEngines();
  } else if(*statusCommand) {
    success = manager.printWorkflowStatus(workflowId);
  } else if (*startWorkflowCommand) {
    success = manager.startWorkflow(workflowFile, workflowEngine);
  } else if (*getWorkflowLogCommand) {
    success = manager.printWorkflowLog(workflowId);
  } else if (*getWorkflowShortcutsCommand) {
    success = manager.printWorkflowShortcuts(workflowId);
  } else if (*getWorkflowInteractionsCommand) {
    success = manager.printWorkflowInteractions(workflowId);
  } else if (*continueWorkflowCommand) {
    success = manager.continueWorkflow(workflowId);
  } else if (*inputInteractionValueCommand) {
    success = manager.inputInteractionValue(workflowId, interactionId, inputValue);
  }

  return success ? 0 : 1;
}
